﻿using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;

using System;
using System.Linq;
using System.Threading.Tasks;
using System.Collections.Generic;

namespace QuickOPC_Leak
{
    public class Connection : IDisposable
    {
        public static ServerElement server = new ServerElement(
                        OpcTechnologies.Com,
                        "opcda:SWToolbox.TOPServer.V6/%7B9877c9f6-80cc-422a-96df-8c958cadda2d%7D",
                        "",
                        new Guid("9877c9f6-80cc-422a-96df-8c958cadda2d"),
                        "9877c9f6-80cc-422a-96df-8c958cadda2d",
                        "SWToolbox.TOPServer.V6",
                        "SWToolbox.TOPServer.V6",
                        "Toolbox OPC Power Server 6.8",
                        "Software Toolbox",
                        new ServerCategories() { OpcDataAccess10 = true, OpcDataAccess20 = true, OpcDataAccess30 = true });

        private EasyDAClient EasyDAClient;
        private const int RWTimeout = 5000;
        private ServerDescriptor ServerDescriptor;

        public bool IsOpen => EasyDAClient != null;

        /// <summary>
        /// Hardcoded test variables.
        /// These are randomly configured by users in the Software, by browsing and mapping tags.
        /// </summary>
        private readonly List<DummyVariable> VariablesToRead = new List<DummyVariable>()
        {
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp4" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp8" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random4" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp4" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp8" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random4" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp4" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Ramp8" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random3" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Random4" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine1" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine2" },
            new DummyVariable() { ID = new Guid(), TagName = "Simulation Examples.Functions.Sine3" },
        };

        public bool Open()
        {
            try
            {
                if (EasyDAClient == null)
                {
                    ServerDescriptor = new ServerDescriptor(server);

                    EasyDAClient = EasyDAClient.Create();
                    EasyDAClient.InstanceParameters.HoldPeriods.ItemDetach = RWTimeout;
                    EasyDAClient.InstanceParameters.HoldPeriods.ServerDetach = RWTimeout;
                    EasyDAClient.InstanceParameters.HoldPeriods.TopicRead = RWTimeout;
                    EasyDAClient.InstanceParameters.HoldPeriods.TopicWrite = RWTimeout;
                    EasyDAClient.InstanceParameters.Timeouts.BrowseAccessPaths = RWTimeout;
                    EasyDAClient.InstanceParameters.Timeouts.BrowseNodes = RWTimeout;
                    EasyDAClient.InstanceParameters.Timeouts.BrowseProperties = RWTimeout;
                    EasyDAClient.InstanceParameters.Timeouts.BrowseServers = RWTimeout;
                    EasyDAClient.InstanceParameters.Timeouts.ReadItem = RWTimeout;
                    EasyDAClient.InstanceParameters.Timeouts.WriteItem = RWTimeout;
                }

                return true;
            }
            catch
            {
                return false;
            }
        }

        public bool Read()
        {
            try
            {
                Task<bool> _t = new Task<bool>(() =>
                {
                    List<DescriptorHolder> Descriptors = new List<DescriptorHolder>();
                    try
                    {                       
                        foreach (var cVariable in VariablesToRead)
                            Descriptors.Add(new DescriptorHolder(cVariable.ID, new DAItemDescriptor(cVariable.TagName)));

                        DAVtqResult[] read_results = null;
                        read_results = EasyDAClient.ReadMultipleItems(ServerDescriptor, Descriptors.Select(d => d.DAItemDescriptor).ToArray());

                        if (read_results == null)
                            return false;

                        if (read_results.Any(result => result.Vtq == null || result.Vtq.Value == null))
                            return false;

                        if (read_results.Length != Descriptors.Count)
                            return false;

                        // Copy read results into local variables.
                        for (int i = 0; i < Descriptors.Count; i++)
                        {
                            var localVariable = VariablesToRead.First(v => v.ID == Descriptors[i].VariableGuid);
                            localVariable.Value = read_results[i].Vtq.Value.ToString();
                            Console.WriteLine($"{localVariable.TagName} - Value: {localVariable.Value}");
                        }

                        return true;
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message);
                        return false;
                    }
                    finally
                    {
                        foreach (var descriptor in Descriptors)
                            descriptor?.Dispose();
                    }
                });

                using (_t)
                {
                    _t.Start();
                    if (_t.Wait(RWTimeout) && _t.Result)
                        return true;
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }

            return false;
        }

        public void Close()
        {
            EasyDAClient?.Dispose();
            EasyDAClient = null;
        }

        public void Dispose()
        {
            Close();
        }
    }
}
